@@ -17,16 +17,17 @@
 
 #include "KdbxXmlReader.h"
 #include "KeePass2RandomStream.h"
-#include "core/Global.h"
-#include "core/Tools.h"
-#include "core/Entry.h"
-#include "core/Group.h"
-#include "core/DatabaseIcons.h"
+#include "core/Clock.h"
 #include "core/Endian.h"
-#include "streams/QtIOCompressor"
+#include "core/Global.h"
+#include "core/Group.h"
+#include "core/Tools.h"
+#include "streams/qtiocompressor.h"
 
-#include <QFile>
 #include <QBuffer>
+#include <QFile>
+
+#define UUID_LENGTH 16
 
 /**
  * @param version KDBX version
@@ -40,9 +41,9 @@ KdbxXmlReader::KdbxXmlReader(quint32 version)
  * @param version KDBX version
  * @param binaryPool binary pool
  */
-KdbxXmlReader::KdbxXmlReader(quint32 version, const QHash<QString, QByteArray>& binaryPool)
+KdbxXmlReader::KdbxXmlReader(quint32 version, QHash<QString, QByteArray> binaryPool)
     : m_kdbxVersion(version)
-    , m_binaryPool(binaryPool)
+    , m_binaryPool(std::move(binaryPool))
 {
 }
 
@@ -52,7 +53,7 @@ KdbxXmlReader::KdbxXmlReader(quint32 version, const QHash<QString, QByteArray>&
  * @param device input file
  * @return pointer to the new database
  */
-Database* KdbxXmlReader::readDatabase(const QString& filename)
+QSharedPointer<Database> KdbxXmlReader::readDatabase(const QString& filename)
 {
     QFile file(filename);
     file.open(QIODevice::ReadOnly);
@@ -65,10 +66,10 @@ Database* KdbxXmlReader::readDatabase(const QString& filename)
  * @param device input device
  * @return pointer to the new database
  */
-Database* KdbxXmlReader::readDatabase(QIODevice* device)
+QSharedPointer<Database> KdbxXmlReader::readDatabase(QIODevice* device)
 {
-    auto db = new Database();
-    readDatabase(device, db);
+    auto db = QSharedPointer<Database>::create();
+    readDatabase(device, db.data());
     return db;
 }
 
@@ -79,7 +80,6 @@ Database* KdbxXmlReader::readDatabase(QIODevice* device)
  * @param db database to read into
  * @param randomStream random stream to use for decryption
  */
-#include "QDebug"
 void KdbxXmlReader::readDatabase(QIODevice* device, Database* db, KeePass2RandomStream* randomStream)
 {
     m_error = false;
@@ -114,17 +114,15 @@ void KdbxXmlReader::readDatabase(QIODevice* device, Database* db, KeePass2Random
     }
 
     if (!m_tmpParent->children().isEmpty()) {
-        qWarning("KdbxXmlReader::readDatabase: found %d invalid group reference(s)",
-                 m_tmpParent->children().size());
+        qWarning("KdbxXmlReader::readDatabase: found %d invalid group reference(s)", m_tmpParent->children().size());
     }
 
     if (!m_tmpParent->entries().isEmpty()) {
-        qWarning("KdbxXmlReader::readDatabase: found %d invalid entry reference(s)",
-                 m_tmpParent->children().size());
+        qWarning("KdbxXmlReader::readDatabase: found %d invalid entry reference(s)", m_tmpParent->children().size());
     }
 
-    const QSet<QString> poolKeys = m_binaryPool.keys().toSet();
-    const QSet<QString> entryKeys = m_binaryMap.keys().toSet();
+    const QSet<QString> poolKeys = Tools::asSet(m_binaryPool.keys());
+    const QSet<QString> entryKeys = Tools::asSet(m_binaryMap.keys());
     const QSet<QString> unmappedKeys = entryKeys - poolKeys;
     const QSet<QString> unusedKeys = poolKeys - entryKeys;
 
@@ -136,7 +134,7 @@ void KdbxXmlReader::readDatabase(QIODevice* device, Database* db, KeePass2Random
         qWarning("KdbxXmlReader::readDatabase: found unused key \"%s\"", qPrintable(key));
     }
 
-    QHash<QString, QPair<Entry*, QString> >::const_iterator i;
+    QMultiHash<QString, QPair<Entry*, QString>>::const_iterator i;
     for (i = m_binaryMap.constBegin(); i != m_binaryMap.constEnd(); ++i) {
         const QPair<Entry*, QString>& target = i.value();
         target.first->attachments()->set(target.second, m_binaryPool[i.key()]);
@@ -144,12 +142,12 @@ void KdbxXmlReader::readDatabase(QIODevice* device, Database* db, KeePass2Random
 
     m_meta->setUpdateDatetime(true);
 
-    QHash<Uuid, Group*>::const_iterator iGroup;
+    QHash<QUuid, Group*>::const_iterator iGroup;
     for (iGroup = m_groups.constBegin(); iGroup != m_groups.constEnd(); ++iGroup) {
         iGroup.value()->setUpdateTimeinfo(true);
     }
 
-    QHash<Uuid, Entry*>::const_iterator iEntry;
+    QHash<QUuid, Entry*>::const_iterator iEntry;
     for (iEntry = m_entries.constBegin(); iEntry != m_entries.constEnd(); ++iEntry) {
         iEntry.value()->setUpdateTimeinfo(true);
 
@@ -179,8 +177,9 @@ QString KdbxXmlReader::errorString() const
 {
     if (m_error) {
         return m_errorStr;
-    }if (m_xml.hasError()) {
-        return QString("XML error:\n%1\nLine %2, column %3")
+    }
+    if (m_xml.hasError()) {
+        return tr("XML error:\n%1\nLine %2, column %3")
             .arg(m_xml.errorString())
             .arg(m_xml.lineNumber())
             .arg(m_xml.columnNumber());
@@ -188,6 +187,11 @@ QString KdbxXmlReader::errorString() const
     return QString();
 }
 
+bool KdbxXmlReader::isTrueValue(const QStringRef& value)
+{
+    return value.compare(QLatin1String("true"), Qt::CaseInsensitive) == 0 || value == "1";
+}
+
 void KdbxXmlReader::raiseError(const QString& errorMessage)
 {
     m_error = true;
@@ -255,7 +259,7 @@ void KdbxXmlReader::parseMeta()
         } else if (m_xml.name() == "Color") {
             m_meta->setColor(readColor());
         } else if (m_xml.name() == "MasterKeyChanged") {
-            m_meta->setMasterKeyChanged(readDateTime());
+            m_meta->setDatabaseKeyChanged(readDateTime());
         } else if (m_xml.name() == "MasterKeyChangeRec") {
             m_meta->setMasterKeyChangeRec(readNumber());
         } else if (m_xml.name() == "MasterKeyChangeForce") {
@@ -342,8 +346,10 @@ void KdbxXmlReader::parseIcon()
 {
     Q_ASSERT(m_xml.isStartElement() && m_xml.name() == "Icon");
 
-    Uuid uuid;
-    QImage icon;
+    QUuid uuid;
+    QByteArray iconData;
+    QString name;
+    QDateTime lastModified;
     bool uuidSet = false;
     bool iconSet = false;
 
@@ -352,15 +358,23 @@ void KdbxXmlReader::parseIcon()
             uuid = readUuid();
             uuidSet = !uuid.isNull();
         } else if (m_xml.name() == "Data") {
-            icon.loadFromData(readBinary());
+            iconData = readBinary();
             iconSet = true;
+        } else if (m_xml.name() == "Name") {
+            name = readString();
+        } else if (m_xml.name() == "LastModificationTime") {
+            lastModified = readDateTime();
         } else {
             skipCurrentElement();
         }
     }
 
     if (uuidSet && iconSet) {
-        m_meta->addCustomIcon(uuid, icon);
+        // Check for duplicate UUID (corruption)
+        if (m_meta->hasCustomIcon(uuid)) {
+            uuid = QUuid::createUuid();
+        }
+        m_meta->addCustomIcon(uuid, iconData, name, lastModified);
         return;
     }
 
@@ -378,19 +392,11 @@ void KdbxXmlReader::parseBinaries()
         }
 
         QXmlStreamAttributes attr = m_xml.attributes();
-
         QString id = attr.value("ID").toString();
-
-        QByteArray data;
-        if (attr.value("Compressed").compare(QLatin1String("True"), Qt::CaseInsensitive) == 0) {
-            data = readCompressedBinary();
-        } else {
-            data = readBinary();
-        }
+        QByteArray data = isTrueValue(attr.value("Compressed")) ? readCompressedBinary() : readBinary();
 
         if (m_binaryPool.contains(id)) {
-            qWarning("KdbxXmlReader::parseBinaries: overwriting binary item \"%s\"",
-                     qPrintable(id));
+            qWarning("KdbxXmlReader::parseBinaries: overwriting binary item \"%s\"", qPrintable(id));
         }
 
         m_binaryPool.insert(id, data);
@@ -415,7 +421,7 @@ void KdbxXmlReader::parseCustomDataItem(CustomData* customData)
     Q_ASSERT(m_xml.isStartElement() && m_xml.name() == "Item");
 
     QString key;
-    QString value;
+    CustomData::CustomDataItem item;
     bool keySet = false;
     bool valueSet = false;
 
@@ -424,15 +430,17 @@ void KdbxXmlReader::parseCustomDataItem(CustomData* customData)
             key = readString();
             keySet = true;
         } else if (m_xml.name() == "Value") {
-            value = readString();
+            item.value = readString();
             valueSet = true;
+        } else if (m_xml.name() == "LastModificationTime") {
+            item.lastModified = readDateTime();
         } else {
             skipCurrentElement();
         }
     }
 
     if (keySet && valueSet) {
-        customData->set(key, value);
+        customData->set(key, item);
         return;
     }
 
@@ -456,8 +464,7 @@ bool KdbxXmlReader::parseRoot()
 
             Group* rootGroup = parseGroup();
             if (rootGroup) {
-                Group* oldRoot = m_db->rootGroup();
-                m_db->setRootGroup(rootGroup);
+                auto oldRoot = m_db->setRootGroup(rootGroup);
                 delete oldRoot;
                 groupParsedSuccessfully = true;
             }
@@ -483,12 +490,12 @@ Group* KdbxXmlReader::parseGroup()
     QList<Entry*> entries;
     while (!m_xml.hasError() && m_xml.readNextStartElement()) {
         if (m_xml.name() == "UUID") {
-            Uuid uuid = readUuid();
+            QUuid uuid = readUuid();
             if (uuid.isNull()) {
                 if (m_strictMode) {
                     raiseError(tr("Null group uuid"));
                 } else {
-                    group->setUuid(Uuid::random());
+                    group->setUuid(QUuid::createUuid());
                 }
             } else {
                 group->setUuid(uuid);
@@ -503,6 +510,10 @@ Group* KdbxXmlReader::parseGroup()
             group->setNotes(readString());
             continue;
         }
+        if (m_xml.name() == "Tags") {
+            group->setTags(readString());
+            continue;
+        }
         if (m_xml.name() == "IconID") {
             int iconId = readNumber();
             if (iconId < 0) {
@@ -510,16 +521,13 @@ Group* KdbxXmlReader::parseGroup()
                     raiseError(tr("Invalid group icon number"));
                 }
                 iconId = 0;
-            } else if (iconId >= DatabaseIcons::IconCount) {
-                qWarning("KdbxXmlReader::parseGroup: icon id \"%d\" not supported", iconId);
-                iconId = DatabaseIcons::IconCount - 1;
             }
 
             group->setIcon(iconId);
             continue;
         }
         if (m_xml.name() == "CustomIconUUID") {
-            Uuid uuid = readUuid();
+            QUuid uuid = readUuid();
             if (!uuid.isNull()) {
                 group->setIcon(uuid);
             }
@@ -587,12 +595,16 @@ Group* KdbxXmlReader::parseGroup()
             parseCustomData(group->customData());
             continue;
         }
+        if (m_xml.name() == "PreviousParentGroup") {
+            group->setPreviousParentGroupUuid(readUuid());
+            continue;
+        }
 
         skipCurrentElement();
     }
 
     if (group->uuid().isNull() && !m_strictMode) {
-        group->setUuid(Uuid::random());
+        group->setUuid(QUuid::createUuid());
     }
 
     if (!group->uuid().isNull()) {
@@ -606,11 +618,11 @@ Group* KdbxXmlReader::parseGroup()
     }
 
     for (Group* child : asConst(children)) {
-        child->setParent(group);
+        child->setParent(group, -1, false);
     }
 
     for (Entry* entry : asConst(entries)) {
-        entry->setGroup(group);
+        entry->setGroup(group, false);
     }
 
     return group;
@@ -637,10 +649,11 @@ void KdbxXmlReader::parseDeletedObject()
 
     while (!m_xml.hasError() && m_xml.readNextStartElement()) {
         if (m_xml.name() == "UUID") {
-            Uuid uuid = readUuid();
+            QUuid uuid = readUuid();
             if (uuid.isNull()) {
                 if (m_strictMode) {
                     raiseError(tr("Null DeleteObject uuid"));
+                    return;
                 }
                 continue;
             }
@@ -675,12 +688,12 @@ Entry* KdbxXmlReader::parseEntry(bool history)
 
     while (!m_xml.hasError() && m_xml.readNextStartElement()) {
         if (m_xml.name() == "UUID") {
-            Uuid uuid = readUuid();
+            QUuid uuid = readUuid();
             if (uuid.isNull()) {
                 if (m_strictMode) {
                     raiseError(tr("Null entry uuid"));
                 } else {
-                    entry->setUuid(Uuid::random());
+                    entry->setUuid(QUuid::createUuid());
                 }
             } else {
                 entry->setUuid(uuid);
@@ -699,7 +712,7 @@ Entry* KdbxXmlReader::parseEntry(bool history)
             continue;
         }
         if (m_xml.name() == "CustomIconUUID") {
-            Uuid uuid = readUuid();
+            QUuid uuid = readUuid();
             if (!uuid.isNull()) {
                 entry->setIcon(uuid);
             }
@@ -729,9 +742,13 @@ Entry* KdbxXmlReader::parseEntry(bool history)
             parseEntryString(entry);
             continue;
         }
+        if (m_xml.name() == "QualityCheck") {
+            entry->setExcludeFromReports(!readBool());
+            continue;
+        }
         if (m_xml.name() == "Binary") {
             QPair<QString, QString> ref = parseEntryBinary(entry);
-            if (!ref.first.isNull() && !ref.second.isNull()) {
+            if (!ref.first.isEmpty() && !ref.second.isEmpty()) {
                 binaryRefs.append(ref);
             }
             continue;
@@ -750,13 +767,24 @@ Entry* KdbxXmlReader::parseEntry(bool history)
         }
         if (m_xml.name() == "CustomData") {
             parseCustomData(entry->customData());
+
+            // Upgrade pre-KDBX-4.1 password report exclude flag
+            if (entry->customData()->contains(CustomData::ExcludeFromReportsLegacy)) {
+                entry->setExcludeFromReports(entry->customData()->value(CustomData::ExcludeFromReportsLegacy)
+                                             == TRUE_STR);
+                entry->customData()->remove(CustomData::ExcludeFromReportsLegacy);
+            }
+            continue;
+        }
+        if (m_xml.name() == "PreviousParentGroup") {
+            entry->setPreviousParentGroupUuid(readUuid());
             continue;
         }
         skipCurrentElement();
     }
 
     if (entry->uuid().isNull() && !m_strictMode) {
-        entry->setUuid(Uuid::random());
+        entry->setUuid(QUuid::createUuid());
     }
 
     if (!entry->uuid().isNull()) {
@@ -787,7 +815,7 @@ Entry* KdbxXmlReader::parseEntry(bool history)
     }
 
     for (const StringPair& ref : asConst(binaryRefs)) {
-        m_binaryMap.insertMulti(ref.first, qMakePair(entry, ref.second));
+        m_binaryMap.insert(ref.first, qMakePair(entry, ref.second));
     }
 
     return entry;
@@ -812,28 +840,9 @@ void KdbxXmlReader::parseEntryString(Entry* entry)
 
         if (m_xml.name() == "Value") {
             QXmlStreamAttributes attr = m_xml.attributes();
-            value = readString();
-
-            bool isProtected = attr.value("Protected") == "True";
-            bool protectInMemory = attr.value("ProtectInMemory") == "True";
-
-            if (isProtected && !value.isEmpty()) {
-                if (m_randomStream) {
-                    QByteArray ciphertext = QByteArray::fromBase64(value.toLatin1());
-                    bool ok;
-                    QByteArray plaintext = m_randomStream->process(ciphertext, &ok);
-                    if (!ok) {
-                        value.clear();
-                        raiseError(m_randomStream->errorString());
-                    } else {
-                        value = QString::fromUtf8(plaintext);
-                    }
-                } else {
-                    raiseError(tr("Unable to decrypt entry string"));
-                    continue;
-                }
-            }
-
+            bool isProtected;
+            bool protectInMemory;
+            value = readString(isProtected, protectInMemory);
             protect = isProtected || protectInMemory;
             valueSet = true;
             continue;
@@ -881,14 +890,6 @@ QPair<QString, QString> KdbxXmlReader::parseEntryBinary(Entry* entry)
             } else {
                 // format compatibility
                 value = readBinary();
-                bool isProtected = attr.hasAttribute("Protected")
-                    && (attr.value("Protected") == "True");
-
-                if (isProtected && !value.isEmpty()) {
-                    if (!m_randomStream->processInPlace(value)) {
-                        raiseError(m_randomStream->errorString());
-                    }
-                }
             }
 
             valueSet = true;
@@ -898,11 +899,13 @@ QPair<QString, QString> KdbxXmlReader::parseEntryBinary(Entry* entry)
     }
 
     if (keySet && valueSet) {
-        if (entry->attachments()->hasKey(key)) {
-            raiseError(tr("Duplicate attachment found"));
-        } else {
-            entry->attachments()->set(key, value);
+        if (entry->attachments()->hasKey(key) && entry->attachments()->value(key) != value) {
+            // NOTE: This only impacts KDBX 3.x databases
+            // Prepend a random string to the key to make it unique and prevent data loss
+            key = key.prepend(QUuid::createUuid().toString().mid(1, 8) + "_");
+            qWarning("Duplicate attachment name found, renamed to: %s", qPrintable(key));
         }
+        entry->attachments()->set(key, value);
     } else {
         raiseError(tr("Entry binary key or value missing"));
     }
@@ -1003,17 +1006,43 @@ TimeInfo KdbxXmlReader::parseTimes()
 
 QString KdbxXmlReader::readString()
 {
-    return m_xml.readElementText();
+    bool isProtected;
+    bool protectInMemory;
+
+    return readString(isProtected, protectInMemory);
+}
+
+QString KdbxXmlReader::readString(bool& isProtected, bool& protectInMemory)
+{
+    QXmlStreamAttributes attr = m_xml.attributes();
+    isProtected = isTrueValue(attr.value("Protected"));
+    protectInMemory = isTrueValue(attr.value("ProtectInMemory"));
+    QString value = m_xml.readElementText();
+
+    if (isProtected && !value.isEmpty()) {
+        QByteArray ciphertext = QByteArray::fromBase64(value.toLatin1());
+        bool ok;
+        QByteArray plaintext = m_randomStream->process(ciphertext, &ok);
+        if (!ok) {
+            value.clear();
+            raiseError(m_randomStream->errorString());
+            return value;
+        }
+
+        value = QString::fromUtf8(plaintext);
+    }
+
+    return value;
 }
 
 bool KdbxXmlReader::readBool()
 {
     QString str = readString();
 
-    if (str.compare("True", Qt::CaseInsensitive) == 0) {
+    if (str.compare("true", Qt::CaseInsensitive) == 0) {
         return true;
     }
-    if (str.compare("False", Qt::CaseInsensitive) == 0) {
+    if (str.compare("false", Qt::CaseInsensitive) == 0) {
         return false;
     }
     if (str.length() == 0) {
@@ -1025,16 +1054,14 @@ bool KdbxXmlReader::readBool()
 
 QDateTime KdbxXmlReader::readDateTime()
 {
-    static QRegularExpression b64regex("^(?:[A-Za-z0-9+/]{4})*(?:[A-Za-z0-9+/]{2}==|[A-Za-z0-9+/]{3}=)?$");
     QString str = readString();
-
-    if (b64regex.match(str).hasMatch()) {
+    if (Tools::isBase64(str.toLatin1())) {
         QByteArray secsBytes = QByteArray::fromBase64(str.toUtf8()).leftJustified(8, '\0', true).left(8);
         qint64 secs = Endian::bytesToSizedInt<quint64>(secsBytes, KeePass2::BYTEORDER);
         return QDateTime(QDate(1, 1, 1), QTime(0, 0, 0, 0), Qt::UTC).addSecs(secs);
     }
 
-    QDateTime dt = QDateTime::fromString(str, Qt::ISODate);
+    QDateTime dt = Clock::parse(str, Qt::ISODate);
     if (dt.isValid()) {
         return dt;
     }
@@ -1043,25 +1070,24 @@ QDateTime KdbxXmlReader::readDateTime()
         raiseError(tr("Invalid date time value"));
     }
 
-    return QDateTime::currentDateTimeUtc();
+    return Clock::currentDateTimeUtc();
 }
 
-QColor KdbxXmlReader::readColor()
+QString KdbxXmlReader::readColor()
 {
     QString colorStr = readString();
 
     if (colorStr.isEmpty()) {
-        return {};
+        return colorStr;
     }
 
     if (colorStr.length() != 7 || colorStr[0] != '#') {
         if (m_strictMode) {
             raiseError(tr("Invalid color value"));
         }
-        return {};
+        return colorStr;
     }
 
-    QColor color;
     for (int i = 0; i <= 2; ++i) {
         QString rgbPartStr = colorStr.mid(1 + 2 * i, 2);
         bool ok;
@@ -1070,19 +1096,11 @@ QColor KdbxXmlReader::readColor()
             if (m_strictMode) {
                 raiseError(tr("Invalid color rgb part"));
             }
-            return {};
-        }
-
-        if (i == 0) {
-            color.setRed(rgbPart);
-        } else if (i == 1) {
-            color.setGreen(rgbPart);
-        } else {
-            color.setBlue(rgbPart);
+            return colorStr;
         }
     }
 
-    return color;
+    return colorStr;
 }
 
 int KdbxXmlReader::readNumber()
@@ -1095,24 +1113,41 @@ int KdbxXmlReader::readNumber()
     return result;
 }
 
-Uuid KdbxXmlReader::readUuid()
+QUuid KdbxXmlReader::readUuid()
 {
     QByteArray uuidBin = readBinary();
     if (uuidBin.isEmpty()) {
-        return {};
+        return QUuid();
     }
-    if (uuidBin.length() != Uuid::Length) {
+    if (uuidBin.length() != UUID_LENGTH) {
         if (m_strictMode) {
             raiseError(tr("Invalid uuid value"));
         }
-        return {};
+        return QUuid();
     }
-    return Uuid(uuidBin);
+    return QUuid::fromRfc4122(uuidBin);
 }
 
 QByteArray KdbxXmlReader::readBinary()
 {
-    return QByteArray::fromBase64(readString().toLatin1());
+    QXmlStreamAttributes attr = m_xml.attributes();
+    bool isProtected = isTrueValue(attr.value("Protected"));
+    QString value = m_xml.readElementText();
+    QByteArray data = QByteArray::fromBase64(value.toLatin1());
+
+    if (isProtected && !data.isEmpty()) {
+        bool ok;
+        QByteArray plaintext = m_randomStream->process(data, &ok);
+        if (!ok) {
+            data.clear();
+            raiseError(m_randomStream->errorString());
+            return data;
+        }
+
+        data = plaintext;
+    }
+
+    return data;
 }
 
 QByteArray KdbxXmlReader::readCompressedBinary()
@@ -1134,7 +1169,7 @@ QByteArray KdbxXmlReader::readCompressedBinary()
     return result;
 }
 
-Group* KdbxXmlReader::getGroup(const Uuid& uuid)
+Group* KdbxXmlReader::getGroup(const QUuid& uuid)
 {
     if (uuid.isNull()) {
         return nullptr;
@@ -1152,7 +1187,7 @@ Group* KdbxXmlReader::getGroup(const Uuid& uuid)
     return group;
 }
 
-Entry* KdbxXmlReader::getEntry(const Uuid& uuid)
+Entry* KdbxXmlReader::getEntry(const QUuid& uuid)
 {
     if (uuid.isNull()) {
         return nullptr;
@@ -1175,4 +1210,3 @@ void KdbxXmlReader::skipCurrentElement()
     qWarning("KdbxXmlReader::skipCurrentElement: skip element \"%s\"", qPrintable(m_xml.name().toString()));
     m_xml.skipCurrentElement();
 }
-
