@@ -19,20 +19,56 @@
 #include "EntryView.h"
 
 #include <QAccessible>
+#include <QDrag>
+#include <QGuiApplication>
 #include <QHeaderView>
-#include <QKeyEvent>
+#include <QListWidget>
 #include <QMenu>
+#include <QPainter>
+#include <QScreen>
 #include <QShortcut>
+#include <QStyledItemDelegate>
+#include <QWindow>
 
+#include "gui/Icons.h"
 #include "gui/SortFilterHideProxyModel.h"
 
+#define ICON_ONLY_SECTION_SIZE 26
+
+class PasswordStrengthItemDelegate : public QStyledItemDelegate
+{
+public:
+    explicit PasswordStrengthItemDelegate(QObject* parent)
+        : QStyledItemDelegate(parent){};
+
+    void initStyleOption(QStyleOptionViewItem* option, const QModelIndex& index) const override
+    {
+        QStyledItemDelegate::initStyleOption(option, index);
+        auto value = index.data(Qt::DecorationRole);
+        if (value.isValid() && value.type() == QVariant::Color && option->rect.width() > 0) {
+            // Rebuild the password strength icon to add a dark border
+            QColor pen(Qt::black);
+            if (option->widget) {
+                pen = option->widget->palette().color(QPalette::Shadow);
+            }
+            auto size = option->decorationSize;
+            QImage image(size.width(), size.height(), QImage::Format_ARGB32_Premultiplied);
+            QPainter p(&image);
+            p.setBrush(value.value<QColor>());
+            p.setPen(pen);
+            p.drawRect(0, 0, size.width() - 1, size.height() - 1);
+            option->icon = QIcon(QPixmap::fromImage(image));
+        }
+    }
+};
+
 EntryView::EntryView(QWidget* parent)
     : QTreeView(parent)
     , m_model(new EntryModel(this))
     , m_sortModel(new SortFilterHideProxyModel(this))
     , m_lastIndex(-1)
     , m_lastOrder(Qt::AscendingOrder)
-    , m_inSearchMode(false)
+    , m_headerMenu(new QMenu(this))
 {
     m_sortModel->setSourceModel(m_model);
     m_sortModel->setDynamicSortFilter(true);
@@ -41,6 +77,7 @@ EntryView::EntryView(QWidget* parent)
     // Use Qt::UserRole as sort role, see EntryModel::data()
     m_sortModel->setSortRole(Qt::UserRole);
     QTreeView::setModel(m_sortModel);
+    QTreeView::setItemDelegateForColumn(EntryModel::PasswordStrength, new PasswordStrengthItemDelegate(this));
 
     setUniformRowHeights(true);
     setRootIsDecorated(false);
@@ -52,24 +89,13 @@ EntryView::EntryView(QWidget* parent)
     // QAbstractItemView::startDrag() uses this property as the default drag action
     setDefaultDropAction(Qt::MoveAction);
 
-    // clang-format off
     connect(this, SIGNAL(doubleClicked(QModelIndex)), SLOT(emitEntryActivated(QModelIndex)));
-    connect(selectionModel(), SIGNAL(selectionChanged(QItemSelection,QItemSelection)), SLOT(emitEntrySelectionChanged()));
-    connect(m_model, SIGNAL(usernamesHiddenChanged()), SIGNAL(viewStateChanged()));
-    connect(m_model, SIGNAL(passwordsHiddenChanged()), SIGNAL(viewStateChanged()));
-    // clang-format on
+    connect(selectionModel(), &QItemSelectionModel::selectionChanged, this, [this] {
+        emit entrySelectionChanged(currentEntry());
+    });
 
     new QShortcut(Qt::CTRL + Qt::Key_F10, this, SLOT(contextMenuShortcutPressed()), nullptr, Qt::WidgetShortcut);
-
-    m_headerMenu = new QMenu(this);
-    m_headerMenu->setTitle(tr("Customize View"));
-    m_headerMenu->addSection(tr("Customize View"));
-
-    m_hideUsernamesAction = m_headerMenu->addAction(tr("Hide Usernames"), this, SLOT(setUsernamesHidden(bool)));
-    m_hideUsernamesAction->setCheckable(true);
-    m_hidePasswordsAction = m_headerMenu->addAction(tr("Hide Passwords"), this, SLOT(setPasswordsHidden(bool)));
-    m_hidePasswordsAction->setCheckable(true);
-    m_headerMenu->addSeparator();
+    new QShortcut(Qt::CTRL + Qt::SHIFT + Qt::Key_J, this, SLOT(jumpToGroupShortcut()), nullptr, Qt::WidgetShortcut);
 
     resetViewToDefaults();
 
@@ -77,25 +103,23 @@ EntryView::EntryView(QWidget* parent)
     // column index as data
     m_columnActions = new QActionGroup(this);
     m_columnActions->setExclusive(false);
-    for (int visualIndex = 1; visualIndex < header()->count(); ++visualIndex) {
+    for (int visualIndex = 0; visualIndex < header()->count(); ++visualIndex) {
         int logicalIndex = header()->logicalIndex(visualIndex);
-        QString caption = m_model->headerData(logicalIndex, Qt::Horizontal, Qt::DisplayRole).toString();
-        if (logicalIndex == EntryModel::Paperclip) {
-            caption = tr("Has attachments", "Entry attachment icon toggle");
-        } else if (logicalIndex == EntryModel::Totp) {
-            caption = tr("Has TOTP", "Entry TOTP icon toggle");
+        auto caption = m_model->headerData(logicalIndex, Qt::Horizontal, Qt::DisplayRole);
+        if (!caption.isValid()) {
+            caption = m_model->headerData(logicalIndex, Qt::Horizontal, Qt::ToolTipRole);
+            if (!caption.isValid()) {
+                // Ignored column, skip it
+                continue;
+            }
         }
 
-        QAction* action = m_headerMenu->addAction(caption);
+        auto action = m_headerMenu->addAction(caption.toString());
         action->setCheckable(true);
         action->setData(logicalIndex);
         m_columnActions->addAction(action);
     }
     connect(m_columnActions, SIGNAL(triggered(QAction*)), this, SLOT(toggleColumnVisibility(QAction*)));
-    connect(header(), &QHeaderView::sortIndicatorChanged, [this](int index, Qt::SortOrder order) {
-        Q_UNUSED(order)
-        header()->setSortIndicatorShown(index != EntryModel::Paperclip && index != EntryModel::Totp);
-    });
 
     m_headerMenu->addSeparator();
     m_headerMenu->addAction(tr("Fit to window"), this, SLOT(fitColumnsToWindow()));
@@ -103,27 +127,15 @@ EntryView::EntryView(QWidget* parent)
     m_headerMenu->addSeparator();
     m_headerMenu->addAction(tr("Reset to defaults"), this, SLOT(resetViewToDefaults()));
 
-    header()->setMinimumSectionSize(24);
     header()->setDefaultSectionSize(100);
     header()->setStretchLastSection(false);
     header()->setContextMenuPolicy(Qt::CustomContextMenu);
 
     connect(header(), SIGNAL(customContextMenuRequested(QPoint)), SLOT(showHeaderMenu(QPoint)));
-    // clang-format off
-    connect(header(), SIGNAL(sectionCountChanged(int,int)), SIGNAL(viewStateChanged()));
-    // clang-format on
-
-    // clang-format off
-    connect(header(), SIGNAL(sectionMoved(int,int,int)), SIGNAL(viewStateChanged()));
-    // clang-format on
-
-    // clang-format off
-    connect(header(), SIGNAL(sectionResized(int,int,int)), SIGNAL(viewStateChanged()));
-    // clang-format on
-
-    // clang-format off
-    connect(header(), SIGNAL(sortIndicatorChanged(int,Qt::SortOrder)), SLOT(sortIndicatorChanged(int,Qt::SortOrder)));
-    // clang-format on
+    connect(header(), SIGNAL(sectionCountChanged(int, int)), SIGNAL(viewStateChanged()));
+    connect(header(), SIGNAL(sectionMoved(int, int, int)), SIGNAL(viewStateChanged()));
+    connect(header(), SIGNAL(sectionResized(int, int, int)), SIGNAL(viewStateChanged()));
+    connect(header(), SIGNAL(sortIndicatorChanged(int, Qt::SortOrder)), SLOT(sortIndicatorChanged(int, Qt::SortOrder)));
 }
 
 void EntryView::contextMenuShortcutPressed()
@@ -151,12 +163,15 @@ void EntryView::sortIndicatorChanged(int logicalIndex, Qt::SortOrder order)
         // do not emit any signals,  header()->setSortIndicator recursively calls this
         // function and the signals are emitted in the else part
     } else {
-        // call emitEntrySelectionChanged even though the selection did not really change
+        // emit entrySelectionChanged even though the selection did not really change
         // this triggers the evaluation of the menu activation and anyway, the position
         // of the selected entry within the widget did change
-        emitEntrySelectionChanged();
+        emit entrySelectionChanged(currentEntry());
         emit viewStateChanged();
     }
+
+    header()->setSortIndicatorShown(true);
+    resetFixedColumns();
 }
 
 void EntryView::keyPressEvent(QKeyEvent* event)
@@ -196,12 +211,6 @@ void EntryView::focusInEvent(QFocusEvent* event)
     QTreeView::focusInEvent(event);
 }
 
-void EntryView::focusOutEvent(QFocusEvent* event)
-{
-    emit entrySelectionChanged(nullptr);
-    QTreeView::focusOutEvent(event);
-}
-
 void EntryView::displayGroup(Group* group)
 {
     m_model->setGroup(group);
@@ -215,11 +224,12 @@ void EntryView::displaySearch(const QList<Entry*>& entries)
     m_model->setEntries(entries);
     header()->showSection(EntryModel::ParentGroup);
 
+    setFirstEntryActive();
+
     // Reset sort column to 'Group', overrides DatabaseWidgetStateSync
     m_sortModel->sort(EntryModel::ParentGroup, Qt::AscendingOrder);
     sortByColumn(EntryModel::ParentGroup, Qt::AscendingOrder);
 
-    setFirstEntryActive();
     m_inSearchMode = true;
 }
 
@@ -249,11 +259,6 @@ void EntryView::emitEntryActivated(const QModelIndex& index)
     emit entryActivated(entry, static_cast<EntryModel::ModelColumn>(m_sortModel->mapToSource(index).column()));
 }
 
-void EntryView::emitEntrySelectionChanged()
-{
-    emit entrySelectionChanged(currentEntry());
-}
-
 void EntryView::setModel(QAbstractItemModel* model)
 {
     Q_UNUSED(model);
@@ -270,6 +275,15 @@ Entry* EntryView::currentEntry()
     }
 }
 
+QList<Entry*> EntryView::selectedEntries()
+{
+    QList<Entry*> list;
+    for (auto row : selectionModel()->selectedRows()) {
+        list.append(m_model->entryFromIndex(m_sortModel->mapToSource(row)));
+    }
+    return list;
+}
+
 int EntryView::numberOfSelectedEntries()
 {
     return selectionModel()->selectedRows().size();
@@ -277,8 +291,11 @@ int EntryView::numberOfSelectedEntries()
 
 void EntryView::setCurrentEntry(Entry* entry)
 {
-    selectionModel()->setCurrentIndex(m_sortModel->mapFromSource(m_model->indexFromEntry(entry)),
-                                      QItemSelectionModel::ClearAndSelect | QItemSelectionModel::Rows);
+    auto index = m_model->indexFromEntry(entry);
+    if (index.isValid()) {
+        selectionModel()->setCurrentIndex(m_sortModel->mapFromSource(index),
+                                          QItemSelectionModel::ClearAndSelect | QItemSelectionModel::Rows);
+    }
 }
 
 Entry* EntryView::entryFromIndex(const QModelIndex& index)
@@ -290,6 +307,11 @@ Entry* EntryView::entryFromIndex(const QModelIndex& index)
     }
 }
 
+QModelIndex EntryView::indexFromEntry(Entry* entry)
+{
+    return m_sortModel->mapFromSource(m_model->indexFromEntry(entry));
+}
+
 int EntryView::currentEntryIndex()
 {
     QModelIndexList list = selectionModel()->selectedRows();
@@ -301,50 +323,6 @@ int EntryView::currentEntryIndex()
     }
 }
 
-/**
- * Get current state of 'Hide Usernames' setting (NOTE: just pass-through for
- * m_model)
- */
-bool EntryView::isUsernamesHidden() const
-{
-    return m_model->isUsernamesHidden();
-}
-
-/**
- * Set state of 'Hide Usernames' setting (NOTE: just pass-through for m_model)
- */
-void EntryView::setUsernamesHidden(bool hide)
-{
-    bool block = m_hideUsernamesAction->signalsBlocked();
-    m_hideUsernamesAction->blockSignals(true);
-    m_hideUsernamesAction->setChecked(hide);
-    m_hideUsernamesAction->blockSignals(block);
-
-    m_model->setUsernamesHidden(hide);
-}
-
-/**
- * Get current state of 'Hide Passwords' setting (NOTE: just pass-through for
- * m_model)
- */
-bool EntryView::isPasswordsHidden() const
-{
-    return m_model->isPasswordsHidden();
-}
-
-/**
- * Set state of 'Hide Passwords' setting (NOTE: just pass-through for m_model)
- */
-void EntryView::setPasswordsHidden(bool hide)
-{
-    bool block = m_hidePasswordsAction->signalsBlocked();
-    m_hidePasswordsAction->blockSignals(true);
-    m_hidePasswordsAction->setChecked(hide);
-    m_hidePasswordsAction->blockSignals(block);
-
-    m_model->setPasswordsHidden(hide);
-}
-
 /**
  * Get current view state
  */
@@ -358,9 +336,12 @@ QByteArray EntryView::viewState() const
  */
 bool EntryView::setViewState(const QByteArray& state)
 {
+    // Reset to unsorted first (https://bugreports.qt.io/browse/QTBUG-86694)
+    header()->setSortIndicator(-1, Qt::AscendingOrder);
     bool status = header()->restoreState(state);
     resetFixedColumns();
     m_columnsNeedRelayout = state.isEmpty();
+    onHeaderChanged();
     return status;
 }
 
@@ -370,8 +351,6 @@ bool EntryView::setViewState(const QByteArray& state)
  */
 void EntryView::showHeaderMenu(const QPoint& position)
 {
-    m_hideUsernamesAction->setChecked(m_model->isUsernamesHidden());
-    m_hidePasswordsAction->setChecked(m_model->isPasswordsHidden());
     const QList<QAction*> actions = m_columnActions->actions();
     for (auto& action : actions) {
         Q_ASSERT(static_cast<QMetaType::Type>(action->data().type()) == QMetaType::Int);
@@ -379,9 +358,9 @@ void EntryView::showHeaderMenu(const QPoint& position)
             continue;
         }
         int columnIndex = action->data().toInt();
-        bool hidden = header()->isSectionHidden(columnIndex) || (header()->sectionSize(columnIndex) == 0);
-        action->setChecked(!hidden);
+        action->setChecked(!isColumnHidden(columnIndex));
     }
+    actions[EntryModel::ParentGroup]->setVisible(inSearchMode());
 
     m_headerMenu->popup(mapToGlobal(position));
 }
@@ -403,11 +382,15 @@ void EntryView::toggleColumnVisibility(QAction* action)
     // least one visible column remains, as the table header will disappear
     // entirely when all columns are hidden
     int columnIndex = action->data().toInt();
+    if (columnIndex == EntryModel::Color) {
+        m_model->setBackgroundColorVisible(!action->isChecked());
+    }
     if (action->isChecked()) {
         header()->showSection(columnIndex);
         if (header()->sectionSize(columnIndex) == 0) {
             header()->resizeSection(columnIndex, header()->defaultSectionSize());
         }
+        resetFixedColumns();
         return;
     }
     if ((header()->count() - header()->hiddenSectionCount()) > 1) {
@@ -456,15 +439,25 @@ void EntryView::fitColumnsToContents()
 }
 
 /**
- * Mark icon-only columns as fixed and resize them to their minimum section size.
+ * Mark icon-only columns as fixed and resize them to icon-only section size
  */
 void EntryView::resetFixedColumns()
 {
-    header()->setSectionResizeMode(EntryModel::Paperclip, QHeaderView::Fixed);
-    header()->resizeSection(EntryModel::Paperclip, header()->minimumSectionSize());
+    for (const auto& col : {EntryModel::Paperclip, EntryModel::Totp, EntryModel::PasswordStrength}) {
+        if (!isColumnHidden(col)) {
+            header()->setSectionResizeMode(col, QHeaderView::Fixed);
 
-    header()->setSectionResizeMode(EntryModel::Totp, QHeaderView::Fixed);
-    header()->resizeSection(EntryModel::Totp, header()->minimumSectionSize());
+            // Increase column width, if sorting, to accommodate icon and arrow
+            auto width = ICON_ONLY_SECTION_SIZE;
+            if (header()->sortIndicatorSection() == col
+                && config()->get(Config::GUI_ApplicationTheme).toString() != "classic") {
+                width += 18;
+            }
+            header()->resizeSection(col, width);
+        }
+    }
+    header()->setMinimumSectionSize(1);
+    header()->resizeSection(EntryModel::Color, ICON_ONLY_SECTION_SIZE);
 }
 
 /**
@@ -472,9 +465,6 @@ void EntryView::resetFixedColumns()
  */
 void EntryView::resetViewToDefaults()
 {
-    m_model->setUsernamesHidden(false);
-    m_model->setPasswordsHidden(true);
-
     // Reduce number of columns that are shown by default
     if (m_inSearchMode) {
         header()->showSection(EntryModel::ParentGroup);
@@ -492,9 +482,12 @@ void EntryView::resetViewToDefaults()
     header()->hideSection(EntryModel::Password);
     header()->hideSection(EntryModel::Expires);
     header()->hideSection(EntryModel::Created);
-    header()->hideSection(EntryModel::Accessed);
     header()->hideSection(EntryModel::Attachments);
     header()->hideSection(EntryModel::Size);
+    header()->hideSection(EntryModel::PasswordStrength);
+    header()->hideSection(EntryModel::Color);
+    header()->hideSection(EntryModel::ParentGroupPath);
+    onHeaderChanged();
 
     // Reset column order to logical indices
     for (int i = 0; i < header()->count(); ++i) {
@@ -522,6 +515,13 @@ void EntryView::resetViewToDefaults()
     }
 }
 
+void EntryView::onHeaderChanged()
+{
+    m_model->setBackgroundColorVisible(isColumnHidden(EntryModel::Color));
+    // Force hide accessed column
+    header()->hideSection(EntryModel::Accessed);
+}
+
 void EntryView::showEvent(QShowEvent* event)
 {
     QTreeView::showEvent(event);
@@ -533,3 +533,85 @@ void EntryView::showEvent(QShowEvent* event)
         m_columnsNeedRelayout = false;
     }
 }
+
+void EntryView::startDrag(Qt::DropActions supportedActions)
+{
+    auto selectedIndexes = selectionModel()->selectedRows(EntryModel::Title);
+    if (selectedIndexes.isEmpty()) {
+        return;
+    }
+
+    // Create a mime data object for the selected rows
+    auto mimeData = m_sortModel->mimeData(selectedIndexes);
+    if (!mimeData) {
+        return;
+    }
+
+    // Create a temporary list widget to display the dragged items
+    int i = 0;
+    QListWidget listWidget;
+    for (auto& index : selectedIndexes) {
+        if (++i > 4) {
+            int remaining = selectedIndexes.size() - i + 1;
+            listWidget.addItem(tr("+ %1 entry(s)...", "", remaining).arg(remaining));
+            break;
+        }
+
+        QIcon icon;
+        icon.addPixmap(m_sortModel->data(index, Qt::DecorationRole).value<QPixmap>());
+
+        auto item = new QListWidgetItem;
+        item->setText(m_sortModel->data(index, Qt::DisplayRole).toString());
+        item->setIcon(icon);
+        listWidget.addItem(item);
+    }
+
+    listWidget.setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
+    listWidget.setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
+    listWidget.setStyleSheet("QListWidget { background-color: palette(highlight); border: 1px solid palette(dark); "
+                             "padding: 4px; color: palette(highlighted-text); }");
+    auto width = listWidget.sizeHintForColumn(0) + 2 * listWidget.frameWidth();
+    auto height = listWidget.sizeHintForRow(0) * listWidget.count() + 2 * listWidget.frameWidth();
+    listWidget.setFixedWidth(width);
+    listWidget.setFixedHeight(height);
+
+    // Grab the screen pixel ratio where the window resides
+    // TODO: Use direct call to screen() when moving to Qt 6
+    auto screen = QGuiApplication::screenAt(window()->geometry().center());
+    if (!screen) {
+        screen = QGuiApplication::primaryScreen();
+    }
+
+    auto pixelRatio = screen->devicePixelRatio();
+
+    // Render the list widget to a pixmap
+    QPixmap pixmap(QSize(width, height) * pixelRatio);
+    pixmap.fill(Qt::transparent);
+    pixmap.setDevicePixelRatio(pixelRatio);
+    listWidget.render(&pixmap);
+
+    // Create a drag object and start the drag
+    auto drag = new QDrag(this);
+    drag->setMimeData(mimeData);
+    drag->setPixmap(pixmap);
+    drag->exec(supportedActions, defaultDropAction());
+}
+
+bool EntryView::isColumnHidden(int logicalIndex)
+{
+    return header()->isSectionHidden(logicalIndex) || header()->sectionSize(logicalIndex) == 0;
+}
+
+void EntryView::jumpToGroupShortcut()
+{
+    // Only allow jump to group in search mode
+    if (!inSearchMode()) {
+        return;
+    }
+
+    auto entry = currentEntry();
+    if (entry) {
+        // Emit the entryActivated signal with ParentGroup column to trigger jump to group
+        emit entryActivated(entry, EntryModel::ParentGroup);
+    }
+}
