@@ -38,6 +38,7 @@ public:
     void setModel(QAbstractItemModel* model) override;
     Entry* currentEntry();
     void setCurrentEntry(Entry* entry);
+    QList<Entry*> selectedEntries();
     Entry* entryFromIndex(const QModelIndex& index);
     QModelIndex indexFromEntry(Entry* entry);
     int currentEntryIndex();
@@ -60,6 +61,7 @@ protected:
     void keyPressEvent(QKeyEvent* event) override;
     void focusInEvent(QFocusEvent* event) override;
     void showEvent(QShowEvent* event) override;
+    void startDrag(Qt::DropActions supportedActions) override;
 
 private slots:
     void emitEntryActivated(const QModelIndex& index);
@@ -70,10 +72,12 @@ private slots:
     void resetViewToDefaults();
     void contextMenuShortcutPressed();
     void sortIndicatorChanged(int logicalIndex, Qt::SortOrder order);
+    void jumpToGroupShortcut();
 
 private:
     void resetFixedColumns();
     bool isColumnHidden(int logicalIndex);
+    void onHeaderChanged();
 
     EntryModel* const m_model;
     SortFilterHideProxyModel* const m_sortModel;
