@@ -32,6 +32,17 @@ private slots:
     void testCopyDataFrom();
     void testClone();
     void testResolveUrl();
+    void testResolveUrlPlaceholders();
+    void testResolveRecursivePlaceholders();
+    void testResolveReferencePlaceholders();
+    void testResolveUuidPlaceholder();
+    void testResolveNonIdPlaceholdersToUuid();
+    void testResolveConversionPlaceholders();
+    void testResolveReplacePlaceholders();
+    void testResolveClonedEntry();
+    void testIsRecycled();
+    void testMoveUpDown();
+    void testPreviousParentGroup();
 };
 
 #endif // KEEPASSX_TESTENTRY_H
